/*
 * A color model that does index mapping for more than 256 colors
 */

import java.awt.*;
import java.awt.image.*;

class MapColorModel extends ColorModel {
    private int rgb[];
    private int mapSize;
    private int offset;

    final static int[] COLORS = {
        0xffc1ffc1, // The various shades of green
        0xff78ff78,
        0xff00d000,
        0xff007000,

        0xff000000,     // black
        0xff0000ff,     // blue
        0xff00ff00,     // green
        0xffff0000,     // red
        0xffffff00,      // yellow
        0xffff00ff,     // magenta
        0xff00ffff,     // cyan
        0xffffffff      // white
    };

    MapColorModel(int offset, long[] data, long min, long range) {
	    super(32);
	    this.offset = offset;
	    mapSize = data.length;
	    rgb = new int[mapSize];
        for (int i = 0; i < mapSize; i++) {
            long val = data[i];
            // Set color for this index based on value
            rgb[i] = (int)( 0xff000000 |
                            ((val - min) * 255 / range) << 16 |
                            (96 - ((val - min) * 96 / range))
                          );
//            System.out.println("Color for index " + i + " is " + rgb[i]);
        }
        System.out.println("Finished generating color model");
    }


    MapColorModel(int offset, long[] data, long[] divisions) {
        super(32);
        this.offset = offset;
        mapSize = data.length;
        rgb = new int[mapSize];

        for (int i = 0; i < mapSize; i++) {
            long val = data[i];
            for (int j = divisions.length -1; j >= 0; j--) {
                if (val >= divisions[j]) {
                    rgb[i] = COLORS[j];
                    break;
                }
            }
        }
        System.out.println("Finished generating color model");
    }



    public int getPixelSize() {
        System.out.println("Someone called getPixelSize()");
        return super.getPixelSize();
    }

    public int getAlpha(int  pixel) {
        if (within(pixel)) return 0xff;	// fully opaque
        else return 0;
    }
    
    /**
     * Returns the red color compoment for the specified pixel in the
     * range 0-255.
     */
    final public int getRed(int pixel) {
        if (within(pixel)) return (rgb[pixel] >> 16) & 0xff;
        else return 0;
    }

    /**
     * Returns the green color compoment for the specified pixel in the
     * range 0-255.
     */
    final public int getGreen(int pixel) {
        if (within(pixel)) return (rgb[pixel-offset] >> 8) & 0xff;
        else return 0;
    }

    /**
     * Returns the blue color compoment for the specified pixel in the
     * range 0-255.
     */
    final public int getBlue(int pixel) {
        if (within(pixel)) return rgb[pixel-offset] & 0xff;
        else return 0;
    }

    public int getRGB(int pixel) {
        if (within(pixel)) return rgb[pixel-offset];
        else return 0;	// Transparent pixel
    }

    private boolean within(int pixel) {
        pixel -= offset;
        return (pixel >= 0 && pixel < mapSize);
    }
    
    /**
     * Returns the size of the color component arrays in this IndexColorModel.
     */
    final public int getMapSize() {
	    return mapSize;
    }


    
    /**
     * Constructs an IndexColorModel from the given arrays of red,
     * green, and blue components.  Pixels described by this color
     * model will all have alpha components of 255 (fully opaque).
     * All of the arrays specifying the color components must have
     * at least the specified number of entries.
     * @param bits	The number of bits each pixel occupies.
     * @param size	The size of the color component arrays.
     * @param r		The array of red color components.
     * @param g		The array of green color components.
     * @param b		The array of blue color components.

    public IndexColorModel(int bits, int size,
			   byte r[], byte g[], byte b[]) {
	super(bits);
	setRGBs(size, r, g, b, null);
    }

    /**
     * Constructs an IndexColorModel from the given arrays of red,
     * green, and blue components.  Pixels described by this color
     * model will all have alpha components of 255 (fully opaque),
     * except for the indicated transparent pixel.  All of the arrays
     * specifying the color components must have at least the specified
     * number of entries.
     * @param bits	The number of bits each pixel occupies.
     * @param size	The size of the color component arrays.
     * @param r		The array of red color components.
     * @param g		The array of green color components.
     * @param b		The array of blue color components.
     * @param trans	The index of the transparent pixel.

    public IndexColorModel(int bits, int size,
			   byte r[], byte g[], byte b[], int trans) {
	super(bits);
	setRGBs(size, r, g, b, null);
	setTransparentPixel(trans);
    }

    /**
     * Constructs an IndexColorModel from the given arrays of red,
     * green, blue and alpha components.  All of the arrays specifying
     * the color components must have at least the specified number
     * of entries.
     * @param bits	The number of bits each pixel occupies.
     * @param size	The size of the color component arrays.
     * @param r		The array of red color components.
     * @param g		The array of green color components.
     * @param b		The array of blue color components.
     * @param a		The array of alpha value components.

    public IndexColorModel(int bits, int size,
			   byte r[], byte g[], byte b[], byte a[]) {
	super(bits);
	if (size > 0 && a == null) {
	    throw new NullPointerException();
	}
	setRGBs(size, r, g, b, a);
    }

    private void setRGBs(int size, byte r[], byte g[], byte b[], byte a[]) {
	map_size = size;
	rgb = new int[Math.max(size, 256)];
	int alpha = 0xff;
	opaque = true;
	for (int i = 0; i < size; i++) {
	    if (a != null) {
		alpha = (a[i] & 0xff);
		if (alpha != 0xff) {
		    opaque = false;
		}
	    }
	    rgb[i] = (alpha << 24)
		| ((r[i] & 0xff) << 16)
		| ((g[i] & 0xff) << 8)
		| (b[i] & 0xff);
	}
    }

    /**
     * Constructs an IndexColorModel from a single arrays of packed
     * red, green, blue and optional alpha components.  The array
     * must have enough values in it to fill all of the needed
     * component arrays of the specified size.
     * @param bits	The number of bits each pixel occupies.
     * @param size	The size of the color component arrays.
     * @param cmap	The array of color components.
     * @param start	The starting offset of the first color component.
     * @param hasalpha	Indicates whether alpha values are contained in
     *			the cmap array.

    public IndexColorModel(int bits, int size, byte cmap[], int start,
			   boolean hasalpha) {
	this(bits, size, cmap, start, hasalpha, -1);
    }

    /**
     * Constructs an IndexColorModel from a single arrays of packed
     * red, green, blue and optional alpha components.  The specified
     * transparent index represents a pixel which will be considered
     * entirely transparent regardless of any alpha value specified
     * for it.  The array must have enough values in it to fill all
     * of the needed component arrays of the specified size.
     * @param bits	The number of bits each pixel occupies.
     * @param size	The size of the color component arrays.
     * @param cmap	The array of color components.
     * @param start	The starting offset of the first color component.
     * @param hasalpha	Indicates whether alpha values are contained in
     *			the cmap array.
     * @param trans	The index of the fully transparent pixel.

    public IndexColorModel(int bits, int size, byte cmap[], int start,
			   boolean hasalpha, int trans) {
	// REMIND: This assumes the ordering: RGB[A]
	super(bits);
	map_size = size;
	rgb = new int[Math.max(size, 256)];
	int j = start;
	int alpha = 0xff;
	opaque = true;
	for (int i = 0; i < size; i++) {
	    rgb[i] = ((cmap[j++] & 0xff) << 16)
		| ((cmap[j++] & 0xff) << 8)
		| (cmap[j++] & 0xff);
	    if (hasalpha) {
		alpha = cmap[j++];
		if (alpha != 0xff) {
		    opaque = false;
		}
	    }
	    rgb[i] |= (alpha << 24);
	}
	setTransparentPixel(trans);
    }
    

    /**
     * Returns the index of the transparent pixel in this IndexColorModel
     * or -1 if there is no transparent pixel.

    final public int getTransparentPixel() {
	return transparent_index;
    }

    /**
     * Copies the array of red color components into the given array.  Only
     * the initial entries of the array as specified by getMapSize() are
     * written.

    final public void getReds(byte r[]) {
	for (int i = 0; i < map_size; i++) {
	    r[i] = (byte) (rgb[i] >> 16);
	}
    }

    /**
     * Copies the array of green color components into the given array.  Only
     * the initial entries of the array as specified by getMapSize() are
     *  written.

    final public void getGreens(byte g[]) {
	for (int i = 0; i < map_size; i++) {
	    g[i] = (byte) (rgb[i] >> 8);
	}
    }

    /**
     * Copies the array of blue color components into the given array.  Only
     * the initial entries of the array as specified by getMapSize() will
     * be written.

    final public void getBlues(byte b[]) {
	for (int i = 0; i < map_size; i++) {
	    b[i] = (byte) rgb[i];
	}
    }

    /**
     * Copies the array of alpha transparency values into the given array.  Only
     * the initial entries of the array as specified by getMapSize() will
     * be written.

    final public void getAlphas(byte a[]) {
	for (int i = 0; i < map_size; i++) {
	    a[i] = (byte) (rgb[i] >> 24);
	}
    }

    private void setTransparentPixel(int trans) {
	if (trans >= map_size || trans < 0) {
	    trans = -1;
	} else {
	    rgb[trans] &= 0x00ffffff;
	    opaque = false;
	}
	transparent_index = trans;
    }


*/
}





