/*
 * An applet to display data as colors on a map.
 * This new version takes the following parameters:
 *    <datatable>  A TableView object containing the data to be displayed
 *    <namevar>    Name of the variable in the table to be used as map
 *                 area names.
 *    <mapfile>    A run-length encoded map file with values corresponding
 *                 to the records in the table
 *    <mapoffset>  The map value corresponding to the record 0 in the table
 */

import java.applet.*;
import java.awt.*;
import java.awt.image.*;
import java.util.*;
import java.net.URL;
import sieve.sluice.*;       
import sieve.sluice.codebook.*;
//import vt.cs.widgets.*;


/**
 * The main class for the Mapper applet.
 * @version 2.0, July 1997
 * @author Jeff Nielsen
 */
public class Mapper extends Applet implements Runnable {

    static final int SCALED_COLORS = 0;
    static final int QUINTILES     = 1;
    static final String[] TYPES = {
        "Scaled Colors",
        "Quintiles"
    };

    // Fields from parameters
    private String imgURL;
    private String dataURL;
    private CodeBookTable m_datatable = null;
    private String m_namevar = "Name";

    private Thread m_setupThread = null;
    
    // applet components
    private Choice m_varSelector = new Choice();  // Variable selector list
    private Choice m_typeSelector = new Choice(); // Map type selection
    private Panel m_controls;
    private MapCanvas m_canvas;
    private Label m_label;          // Displays messages to user
    private Label m_info;           // Used in place of the status box
    private Frame m_parent;         // Used for setting the cursor

    // FLAGS
    private boolean m_dataReady = false;

    // Fields to record the state of things
//    private int m_numRecords = 0;               // # of records in DB
    private String[] m_recordNames;             // Names of records
    private Vector m_varNames;                  // All available fields
    private Object[] m_currVarValues;             // Array that holds the values

    /** 
     * Constructor
     */
    public Mapper() {}

    public String getAppletInfo() {
        return "Name: Mapper\r\n" + "Author: Jeff Nielsen";
    }

    /**
     * Returns an array of strings describing the parameters understood
     * by this applet.
     * Mapper Parameter Information: { "Name", "Type", "Description" },
     */
    public String[][] getParameterInfo() {
        String[][] info =
            {
            };
        return info;            
    }

    /**
     * Initialize the applet
     */
    public void init()  {
        try {
            // Required parameters
            imgURL = getParameter("mapfile");
            dataURL = getParameter("datatable");
            // Optional parameters
            String param = getParameter("namevar");
            if (param != null) m_namevar = param;
            int mapOffset = 0;
            param = getParameter("mapoffset");
            if (param != null) mapOffset = Integer.parseInt(param);

            
            // Add the variable list and the output area
            setLayout(new BorderLayout());

            for (int i = 0; i < TYPES.length; i++)
                m_typeSelector.addItem(TYPES[i]);
            m_varSelector.addItem("NO VARIABLE SELECTED");
            m_controls = new Panel();
            m_controls.setLayout(new FlowLayout(FlowLayout.LEFT));
            m_controls.add(m_typeSelector);
            m_controls.add(m_varSelector);
            add("North", m_controls);
            m_label = new Label(
                "Please wait while the data file is accessed . . .",
                Label.CENTER);
            add("South", m_label);

            // Change to "waiting" mouse pointer--find parent
            Container c = getParent();
            while (! (c instanceof Frame)) c = c.getParent();
            m_parent = (Frame)c;
            m_parent.setCursor(Frame.WAIT_CURSOR);


            // Now add the map canvas, created from the mapfile.
            m_canvas = new MapCanvas(imgURL, mapOffset, this);

            // Now figure out whether to put canvas in a RadarPane
            Dimension aSize = size(), cSize = m_canvas.minimumSize();
/*
            if (cSize.height > aSize.height || cSize.width > aSize.width) {
                rp = new RadarPane();
                rp.add(m_canvas);
                add("Center", rp);
            }
            else
*/
                add("Center", m_canvas);

        } catch (Exception e) {
            e.printStackTrace();
        }
    }

/*
    public void showStatus(String msg) {
        m_info.setText(msg);
    }
*/
    /**
     * Initializes the interchange object to access the data
     * Runs in its own thread so that users don't have to wait
     * for this to finish before they see something on the screen.
     */
    public void run() {
        Choice temp = new Choice();
        m_varNames = new Vector();

        try {
            m_datatable = new SGMLTable(dataURL);
            // Check to make sure we got something

            // Store area names
	        Object[] elems = m_datatable.get(m_namevar);
	        m_recordNames = new String[elems.length];
	        for (int i = 0; i < elems.length; i++ )
		    m_recordNames[i] = (String)elems[i];

            // Get rest of variable names
            String[] varNames = m_datatable.fieldNames();
            for (int i = 0; i < varNames.length; i++) {
                // Make sure it's a variable we want
                String name = varNames[i];
                if (name.equals(m_namevar)) continue;
                FieldDescriptor fd = m_datatable.getFieldDescriptor(name);
                if (fd.getLevel() < FieldDescriptor.LEVEL_INTERVAL) continue;
                
                // Store real names in a vector in same order as in list
                temp.addItem(fd.getDescription());
                m_varNames.addElement(name);
            }
        } catch (Exception e) {
            e.printStackTrace();
        }

        m_controls.remove(m_varSelector); // Now remove the old one from the 
        m_varSelector = temp;             // layout manager & add the new one
        m_controls.add(m_varSelector);
        validate();                 // I do need this.

        // Display the first variable
        changeVariable(true);
    }

    public void paint(Graphics g) {   // Nothing to paint now with MapCanvas
//        System.out.println("paint called");
    }

    /**
     * Called when the page containing the applet
     * first appears on the screen. I'm using this to start the thread that
     * loads the data.
     */
    public void start() {
        if (m_setupThread == null) {
            m_setupThread = new Thread(this);
            m_setupThread.start();
        }
    }

    /** 
     * Handles the event of choosing a variable in the selection
     * list.
     */
    public boolean action(Event evt, Object what) {
        if (evt.target == m_varSelector) {
            changeVariable(true);
            return true;
        }
        else if (evt.target == m_typeSelector) {
            changeVariable(false);
            return true;
        }
        return false;
    }

    
    /**
     * Called to update the map, etc. when a new variable is chosen from 
     * the list.
     */
    private void changeVariable(boolean newData) {
        output("Please wait while the new map is generated . . .");
        m_parent.setCursor(Frame.WAIT_CURSOR);
        if (newData) {
            m_dataReady = false;
            String currVariable = (String)
                m_varNames.elementAt(m_varSelector.getSelectedIndex());
            try {
                m_currVarValues = m_datatable.get(currVariable);
            } catch (Exception e) {
                e.printStackTrace();
            }
            m_dataReady = true;
        }
        m_canvas.colorMap(m_currVarValues, m_typeSelector.getSelectedIndex());
    }

    /**
     * Prints a message at the bottom of the window
     * @param s The message
     */
    private void output(String s) {
        m_label.setText(s);
    }

    /**
     * Called by the MapCanvas to indicate that the mouse is over a 
     * particular area.  This method prints the pertinent information
     * in the status line
     */
    public void showInfo(int area) {
        try {
            if (m_dataReady)
                showStatus(m_recordNames[area] + ": " + m_currVarValues[area]);
            else showStatus(m_recordNames[area]);
        } catch (ArrayIndexOutOfBoundsException e) { 
            showStatus("");  // Don't show anything if not over a valid state
        }
    }

    /**
     * Called by the MapCanvas to indicate that the map has been (re)drawn
     */
    public void redrawDone() {
        if (m_dataReady) {  // Don't show message with first (black) map
            output("Choose a different variable to recolor the map.");
            m_parent.setCursor(Frame.DEFAULT_CURSOR);
        }
    }

}


class MapCanvas extends Canvas {
    private RLImage m_rli;
    private int m_mapOffset;
    private Mapper m_parent;

    private Image m_img = null;
    boolean colored = false, cleared = false;
    int m_xoffset = 0, m_yoffset = 0;

    MapCanvas(String imgURL, int mapoffset, Mapper parent) 
        throws java.net.MalformedURLException, java.io.IOException {
        m_rli = new RLImage(new URL(imgURL).openStream(), 
            // we should probably get some good default colors for the image
            // but for now we'll just make it all black
            new SingleColorModel(mapoffset, 0xff000000),
            true);
        m_mapOffset = mapoffset;
        m_parent = parent;
        m_img = createImage(m_rli);
    }

    public void colorMap(Object[] data, int mapType) {
        long[] temp = new long[data.length];
        switch(mapType) {
          case Mapper.SCALED_COLORS:
            long min = Long.MAX_VALUE, max = 0; 
                // Calculate the max & min
            for (int i = 0; i < temp.length; i++) {
                long val = ((Number)data[i]).longValue(); 
                if (val < min) min = val;
                if (val > max) max = val;
                temp[i] = val;
            }
            long range = max - min;
            if (range == 0) range = 1;   // hack, fix this later
            m_rli.changeColorModel(new MapColorModel(
    	        m_mapOffset, temp, min, range));
            break;

          case Mapper.QUINTILES:
            int n = temp.length;
            for (int i = 0; i < n; i++)
                temp[i] = ((Number)data[i]).longValue();
            long[] sorted = new long[n];
            System.arraycopy(temp, 0, sorted, 0, n);
            Sort.quicksort(sorted, 0, n-1);

            long[] quintiles = new long[5];
            for (int i = 0; i < 5; i++) {
                int index = (int)(n * (i / 5.0));
                quintiles[i] = sorted[index];
                System.out.println(quintiles[i]);
            }
            m_rli.changeColorModel(new MapColorModel(
                m_mapOffset, temp, quintiles));
            break;
              
        }
        colored = true;     // Means this has been called at least once

    }


    public Dimension minimumSize() { 
        return new Dimension(m_rli.width,m_rli.height);
    }
        
    public Dimension preferredSize() { return minimumSize(); }

    public void printAll(Graphics g) {
        System.out.println("printAll in MapCanvas called.");
        super.printAll(g);
    }

    public void paint(Graphics g) { 
//        System.out.println("paint in MapCanvas called.");
        // Figure out where to put the image so it's centered
        Dimension d = size();
        m_xoffset = (d.width - m_rli.width) / 2;
        m_yoffset = (d.height - m_rli.height) / 2;
        g.drawImage(m_img, m_xoffset, m_yoffset, this);
    }

    /**
     * Overrides Component.update so that it doesn't clear the
     * background before calling paint().  Makes for less flickering.
     */ 
    public void update(Graphics g) {   
        if (colored && !cleared) {      // Should clear the screen once.
            cleared = true;
            System.out.println("OK, clearing now");
            super.update(g);
        }
        else paint(g); 
    }
    
    public boolean mouseMove(Event evt, int x, int y) {
        x -= m_xoffset;
        y -= m_yoffset;
        try {
            m_parent.showInfo(m_rli.getPixel(x,y) - m_mapOffset);
            return true;
        } catch (Exception e) {
            // Outside of image
        }
        return false;
    }

    /**
     * Notified about the status of the new image
     */
    public boolean imageUpdate(Image img, int infoflags, 
        int x, int y, int width, int height) {
        if ((infoflags & FRAMEBITS ) != 0) {
            m_parent.redrawDone();
        }
        return super.imageUpdate(img, infoflags, x, y, width, height);
    }

}
