import java.io.*;
import java.awt.image.*;
import java.util.*;

public class RLImage implements ImageProducer {

    private static final boolean verbose = false;
    private static final int MAX_PIXELS = 100000; // Max # of pixels to buffer
                                                  // (4 bytes each)
    static final int TAG4 = 0x84; 
    static final int TAG8 = 0x88; 
    static final int TAG12 = 0x8c; 
    static final int TAG16 = 0x90; 

    int tag;                    // Kind of run-length image--one of above vals.
    byte runsize;               // Number of bytes in a run--1, 2, or 3
    int width;                  // Width of image
    int height;                 // Height of image
    private byte[] runs;        // Stores the image data
    private int[] rowVector;    // Stores offsets in runs[] for each row of the image

    ColorModel model;           // Model used to decode image
    ColorModel sendModel;       // Model sent to ImageConsumer
    boolean littleEndian;       // Whether or not image byte ordering is little endian
    Hashtable properties = new Hashtable();
    Vector consumers = new Vector();

    boolean changing;           // Whether the colors of an image can change


    RLImage(InputStream is) throws IOException {
        initialize(is, new DirectColorModel(16, 0xf0, 0xf, 0xf00), false, null);
    }

    RLImage(InputStream is, ColorModel cm) throws IOException {
        initialize(is, cm, false, null);
    }

    RLImage(InputStream is, boolean change) throws IOException {
        initialize(is, new DirectColorModel(16, 0xf0, 0xf, 0xf00), change, null);
    }

    RLImage(InputStream is, ColorModel cm, boolean change) throws IOException {
        initialize(is, cm, change, null);
    }

    private void initialize(InputStream is, ColorModel cm, boolean changeable,
        Hashtable props) throws IOException {
        model = cm;
        sendModel = (model instanceof DirectColorModel 
                    ? model 
                    : ColorModel.getRGBdefault());
            
        changing = changeable;
        if (props == null) props = new Hashtable();
        properties = props;

        // Convert the stream to an object
        DataInput input;
        byte[] buf = new byte[3];
        is.read(buf);

            // Check header
        if ((buf[0] == 'R') || (buf[1] == 'U') || (buf[2] == 'N')) {
            littleEndian = true;
            input = new LEDataInputStream(is);  // Little-endian byte ordering
        }
        else if ((buf[0] == 'R') || (buf[1] == 'L') || (buf[2] == 'I')) {
            littleEndian = false;
            input = new DataInputStream(is);    // Big-endian byte ordering
        }
        else throw new IOException("Not a run-length encoded file");

        // Get tag, width, & height
        tag = input.readUnsignedByte();
        switch (tag) {
          case TAG4: runsize = 1; break;
          case TAG8:
          case TAG12: runsize = 2; break;
          case TAG16: runsize = 3; break;
          default: throw new IOException("Unknown tag type");
        }
        height = input.readUnsignedShort();
        width = input.readUnsignedShort();

        // Store row vector, converted to actual offsets
        // Last element in row vector is length of data
        rowVector = new int[height + 1];
        if (input.readUnsignedShort() != 8 + 2 * (height + 1))
            System.out.println("First run in unexpected place!");
        rowVector[0] = 0;
        for (int i = 1; i < rowVector.length; i++) {
            rowVector[i] = input.readUnsignedShort() + rowVector[i-1];
        }

        // Now dump run data into a big byte array
        runs = new byte[rowVector[height]];
        input.readFully(runs);
    }

    /**
     * Give access to the runs array as a DataInput object.
     * If image has little-endian byte ordering, give them an LEDataInputStream
     */
    DataInput getDataInputStream() {
        if (littleEndian) 
            return new LEDataInputStream(new ByteArrayInputStream(runs));
        else
            return new DataInputStream(new ByteArrayInputStream(runs));
    }


    /**
     * Get the pixel value at the specified coordinates
     */
    public int getPixel(int x, int y) throws Exception {
        if (x >= width || x < 0 || y >= height || y < 0)
            throw new Exception("Invalid coordinates");

        int offset = rowVector[y];      // xhere row y starts in the runs
        RLRun next = getRun(offset);
        int currCol = next.length - 1;  // last column in this run
        while (x > currCol) {
            offset += runsize;
            next = getRun(offset);
            currCol += next.length;
        }
        return next.val;
    }

    /**
     * Decode the run from the specified offset in the runs array
     */
    public RLRun getRun(int off) {
        int length = 0, val = 0, byte1, byte2, byte3;

        switch (tag) {
          case TAG4:
            byte1 = runs[off] & 0xff;
            length = byte1 >>> 4;
            val = byte1 & 0xf;
            if (length == 0) length = 16;
            break;

          case TAG8:        // This one's backwards--val before length!!
            byte1 = runs[off] & 0xff;
            byte2 = runs[off+1] & 0xff;
            if (littleEndian) { length = byte1; val = byte2; }
            else {length = byte2; val = byte1; }
            if (length == 0) length = 256;
            break;

          case TAG12:
            byte1 = runs[off] & 0xff;
            byte2 = runs[off+1] & 0xff;
            if (littleEndian) {
                length = byte2 >>> 4;
                val = byte1 | (byte2 & 0xf) << 8;
            }
            else {
                length = byte1 >>> 4;
                val = byte2 | (byte1 & 0xff) << 8;
            }
            if (length == 0) length = 16;
            break;

          case TAG16:
            byte1 = runs[off] & 0xff;
            byte2 = runs[off+1] & 0xff;
            byte3 = runs[off+2] & 0xff;
            length = byte1;
            if (littleEndian) val = byte2 | (byte3 << 8);
            else val = byte3 | (byte2 << 8);
            if (length == 0) length = 256;
            break;
        }
        return new RLRun(length, val);
    }


    public String toString() {
        return ("Tag: " + tag + "  Width: " + width + "  Height: " + height + 
            "  Data: " + runs.length + " bytes.");
    }


    // ImageProducer stuff follows //////////////////////////////////////////

    /**
     * Adds an ImageConsumer to the list of consumers interested in
     * data for this image.
     */
    public synchronized void addConsumer(ImageConsumer ic) {
        System.out.println("addConsumer called for " + ic);
        if (consumers.contains(ic)) return;
            
        consumers.addElement(ic);
        try {
            initConsumer(ic);
            produceImage(ic);
            if (isConsumer(ic)) {
                ic.imageComplete(changing
                    ? ImageConsumer.SINGLEFRAMEDONE
                    : ImageConsumer.STATICIMAGEDONE);
                if (!changing && isConsumer(ic)) {
                    ic.imageComplete(ImageConsumer.IMAGEERROR);
                    removeConsumer(ic);
                }
            }
        } catch (Exception e) {
            if (isConsumer(ic)) ic.imageComplete(ImageConsumer.IMAGEERROR);
        }
    }

    public synchronized boolean isConsumer(ImageConsumer ic) {
        return consumers.contains(ic);
    }

    public synchronized void removeConsumer(ImageConsumer ic) {
        System.out.println("removeConsumer called for " + ic);
        consumers.removeElement(ic);
    }

    public void startProduction(ImageConsumer ic) {
        System.out.println("startProduction called for " + ic);
        addConsumer(ic);
    }

    public void requestTopDownLeftRightResend(ImageConsumer ic) {
        System.out.println("requestTDLRReset called for " + ic);
        // ignore for now
    }

    /**
     * Send a recolored image to any
     * ImageConsumers that are currently interested in the data for
     * this image.
     */
    public synchronized void changeColorModel(ColorModel cm) {
        model = cm; 
        sendModel = (model instanceof DirectColorModel 
                    ? model 
                    : ColorModel.getRGBdefault());

        if (changing) {
            Enumeration enum = consumers.elements();
            while (enum.hasMoreElements()) {
                ImageConsumer ic = (ImageConsumer) enum.nextElement();
                produceImage(ic);
                if (isConsumer(ic)) 
                    ic.imageComplete(ImageConsumer.SINGLEFRAMEDONE);
            }
        }
    }

    private void initConsumer(ImageConsumer ic) {
        if (isConsumer(ic)) {
            ic.setDimensions(width, height);
            ic.setProperties(properties);  
            ic.setColorModel(sendModel);
            ic.setHints(changing
                ? (ImageConsumer.TOPDOWNLEFTRIGHT |
                   ImageConsumer.COMPLETESCANLINES)
                : (ImageConsumer.TOPDOWNLEFTRIGHT |
                   ImageConsumer.COMPLETESCANLINES |
                   ImageConsumer.SINGLEPASS |
                   ImageConsumer.SINGLEFRAME));
        }
    }


    /**
     * Produce image for an ImageConsumer
     */
    private void produceImage(ImageConsumer ic) {
        // Figure out how many lines to send at once
        int numRows = MAX_PIXELS / width;
        if (numRows == 0) numRows = 1;
        int rowCount = 0;   // # of rows currently in buffer
        int[] pixels = new int[width * numRows];

         // Parse the data in the runs array & send pixel values
        int byteOffset = 0;
        for (int row = 0; row < height; row++) {  // for each row
            for (int col = 0; col < width; ) {    // for each column
                RLRun r = getRun(byteOffset);
                // Could do color mapping right here and always send
                // the RGBDefault color model below
                // Chose to do it this way, because 1.02 won't handle my color models!
                if (! (model instanceof DirectColorModel))
                    r.val = model.getRGB(r.val);
                
                for (int j = 0; j < r.length; j++) 
                    pixels[rowCount * width + col + j] = r.val;
                col += r.length;
                byteOffset += runsize;
            }   // end for each column

            // Send pixels when buffer is full
            if (++rowCount == numRows) {
                ic.setPixels(0, row - rowCount + 1, width, numRows, 
                    sendModel, pixels, 0, width);
                    rowCount = 0;
                System.out.println("Sent a block of pixels.");
                }
         }   // end for each row

        // send any remaining buffered pixels, then signal finished
        ic.setPixels(0, height - rowCount, width,
            rowCount, sendModel, pixels, 0, width);
        System.out.println("Finished delivering pixels.");
    }
}

///////////////////////////////////////////////////////////////////

/**
 * A Run object stores the length & value of a run
 */
class RLRun{
    int length;
    int val;

    RLRun(int length, int val) { this.length = length; this.val = val; }
}
